@@ -3,33 +3,38 @@ from __future__ import annotations
 
 import asyncio
 import logging
-from typing import Callable, Dict, List, Optional, Tuple, Union
+from types import TracebackType
+from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
 
 from .client import SlimClient
-from .const import EventType
+from .const import EventType, SlimEvent
 from .discovery import start_discovery
+from .json_rpc import SlimJSONRPC
 
-EventCallBackType = Callable[[EventType, SlimClient], None]
+EventCallBackType = Callable[[SlimEvent], None]
 EventSubscriptionType = Tuple[EventCallBackType, Tuple[EventType], Tuple[str]]
 
 
 class SlimServer:
     """Server holding the SLIMproto players."""
 
-    def __init__(self, port: int = 3483, json_port: Optional[int] = None) -> None:
+    def __init__(self, port: int = 3483, json_port: Optional[int] = 3484) -> None:
         """
         Initialize SlimServer instance.
 
         control_port: The TCP port for the slimproto communication, default is 3483.
         json_port: Optionally start a simple json rpc server on this port for compatability
-        with players relying on the server providing this feature.
+        with players relying on the server providing this feature. None to disable.
         """
         self.logger = logging.getLogger(__name__)
         self.port = port
         self.json_port = json_port
         self._subscribers: List[EventSubscriptionType] = []
-        self._bgtasks: List[asyncio.Task] = []
+        self._socket_servers: List[Union[asyncio.Server, asyncio.BaseTransport]] = []
         self._players: Dict[str, SlimClient] = {}
+        self._jsonrpc: Optional[SlimJSONRPC] = None
+        if json_port is not None:
+            self._jsonrpc = SlimJSONRPC(self, json_port)
 
     @property
     def players(self) -> List[SlimClient]:
@@ -43,23 +48,24 @@ class SlimServer:
     async def start(self):
         """Start running the server."""
         self.logger.info("Starting SLIMProto server on port %s", self.port)
-        self._bgtasks = [
+        self._socket_servers = [
             # start slimproto server
-            asyncio.create_task(
-                asyncio.start_server(self._create_client, "0.0.0.0", self.port)
-            ),
+            await asyncio.start_server(self._create_client, "0.0.0.0", self.port),
             # setup discovery
-            asyncio.create_task(start_discovery(self.port, self.json_port)),
+            await start_discovery(self.port, self.json_port),
         ]
+        if self._jsonrpc is not None:
+            self._socket_servers.append(await self._jsonrpc.start())
 
     async def stop(self):
-        """Stop runninbg the server."""
-        for task in self._bgtasks:
-            task.cancel()
+        """Stop running the server."""
+        for client in list(self._players.values()):
+            client.disconnect()
+        self._players = {}
+        for _server in self._socket_servers:
+            _server.close()
 
-    def signal_event(
-        self, event_type: EventType, player: Optional[SlimClient] = None
-    ) -> None:
+    def signal_event(self, event: SlimEvent) -> None:
         """
         Signal event to all subscribers.
 
@@ -67,16 +73,18 @@ class SlimServer:
             :param event_details: optional details to send with the event.
         """
         for cb_func, event_filter, player_filter in self._subscribers:
-            if player and player_filter and player.id not in player_filter:
+            if (
+                event.player_id
+                and player_filter
+                and event.player_id not in player_filter
+            ):
                 continue
-            if event_filter and event_type not in event_filter:
+            if event_filter and event.type not in event_filter:
                 continue
             if asyncio.iscoroutinefunction(cb_func):
-                asyncio.create_task(cb_func(event_type, player))
+                asyncio.create_task(cb_func(event))
             else:
-                asyncio.get_running_loop().call_soon_threadsafe(
-                    cb_func, event_type, player
-                )
+                asyncio.get_running_loop().call_soon_threadsafe(cb_func, event)
 
     def subscribe(
         self,
@@ -115,27 +123,37 @@ class SlimServer:
         """Create player from new connection on the socket."""
         addr = writer.get_extra_info("peername")
         self.logger.debug("Socket client connected: %s", addr)
-        SlimClient(reader, writer, self._client_callback)
 
-    def _client_callback(self, event: EventType, client: SlimClient):
-        """Handle event/callback from SlimClient."""
-        player_id = client.player_id
-        if not player_id:
-            return
+        def client_callback(
+            event_type: EventType, client: SlimClient, data: Any = None
+        ):
+            player_id = client.player_id
 
-        if event == EventType.PLAYER_DISCONNECTED:
-            # player disconnected, cleanup and emit event
-            player = self._players.pop(player_id)
-            player.disconnect()
-            self.signal_event(EventType.PLAYER_REMOVED, player)
-            return
+            if event_type == EventType.PLAYER_DISCONNECTED:
+                self._players.pop(player_id, None)
 
-        if event == EventType.PLAYER_CONNECTED:
-            # new player (or reconnected)
-            assert player_id not in self._players
-            self._players[player_id] = client
-            self.signal_event(EventType.PLAYER_ADDED, client)
-            return
+            if event_type == EventType.PLAYER_CONNECTED:
+                assert player_id not in self._players
+                self._players[player_id] = client
 
-        # forward all other events
-        self.signal_event(event, client)
+            # forward all other events as-is
+            self.signal_event(SlimEvent(event_type, player_id, data))
+
+        SlimClient(reader, writer, client_callback)
+
+    async def __aenter__(self) -> "SlimServer":
+        """Return Context manager."""
+        await self.start()
+        return self
+
+    async def __aexit__(
+        self,
+        exc_type: Type[BaseException],
+        exc_val: BaseException,
+        exc_tb: TracebackType,
+    ) -> Optional[bool]:
+        """Exit context manager."""
+        await self.stop()
+        if exc_val:
+            raise exc_val
+        return exc_type
